@@ -6,7 +6,8 @@ import numpy as np
 
 from exetera.core import persistence as pers
 from exetera.core.data_writer import DataWriter
-
+from exetera.core import operations as ops
+from exetera.core.utils import Timer
 
 class Reader:
     def __init__(self, field):
@@ -25,10 +26,12 @@ class IndexedStringReader(Reader):
             raise ValueError(error.format(field, fieldtype))
         self.chunksize = field.attrs['chunksize']
         self.datastore = datastore
+        
 
     def __getitem__(self, item):
         try:
             if isinstance(item, slice):
+
                 start = item.start if item.start is not None else 0
                 stop = item.stop if item.stop is not None else len(self.field['index']) - 1
                 step = item.step
@@ -226,13 +229,17 @@ class IndexedStringWriter(Writer):
         self.fieldtype = fieldtype
         self.timestamp = timestamp
         self.datastore = datastore
+        self.chunk_accumulated = 0
 
-        self.values = np.zeros(self.datastore.chunksize, dtype=np.uint8)
-        self.indices = np.zeros(self.datastore.chunksize, dtype=np.int64)
-        self.ever_written = False
-        self.accumulated = 0
-        self.value_index = 0
-        self.index_index = 0
+        # self.values = np.zeros(self.datastore.chunksize, dtype=np.uint8)
+        # self.indices = np.zeros(self.datastore.chunksize, dtype=np.int64)
+        # self.ever_written = False
+        # self.accumulated = 0
+        # self.value_index = 0
+        # self.index_index = 0
+        # self.chunk_accumulated = 0
+        if 'index' not in self.field.keys():
+            DataWriter.write(self.field, 'index', [0], 1)
 
     def chunk_factory(self, length):
         return [None] * length
@@ -240,36 +247,67 @@ class IndexedStringWriter(Writer):
     def write_part(self, values):
         """
         Writes a list of strings in indexed string form to a field.
-        
+
         :param values: a list of utf8 strings
         """
-        if not self.ever_written:
-            self.indices[0] = self.accumulated
-            self.index_index = 1
-            self.ever_written = True
+        # if not self.ever_written:
+        #     self.indices[0] = self.accumulated
+        #     self.index_index = 1
+        #     self.ever_written = True
+        #
+        # for s in values:
+        #     if isinstance(s, str):
+        #         evalue = s.encode()
+        #     else:
+        #         evalue = s
+        #
+        #     for v in evalue:
+        #         self.values[self.value_index] = v
+        #         self.value_index += 1
+        #         if self.value_index == self.datastore.chunksize:
+        #             DataWriter.write(self.field, 'values', self.values, self.value_index)
+        #             self.value_index = 0
+        #         self.accumulated += 1
+        #     self.indices[self.index_index] = self.accumulated
+        #     self.index_index += 1
+        #     if self.index_index == self.datastore.chunksize:
+        #         DataWriter.write(self.field, 'index', self.indices, self.index_index)
+        #         self.index_index = 0
 
+        accumulated = self.field['index'][-1]
+        indices = np.zeros(len(values), dtype=np.int64)
+        chars = np.zeros(self.chunksize, dtype=np.uint8)
+        index_index = 0
+        char_index = 0
         for s in values:
-            evalue = s.encode()
+            if isinstance(s, str):
+                evalue = s.encode()
+            else:
+                evalue = s
+
             for v in evalue:
-                self.values[self.value_index] = v
-                self.value_index += 1
-                if self.value_index == self.datastore.chunksize:
-                    DataWriter.write(self.field, 'values', self.values, self.value_index)
-                    self.value_index = 0
-                self.accumulated += 1
-            self.indices[self.index_index] = self.accumulated
-            self.index_index += 1
-            if self.index_index == self.datastore.chunksize:
-                DataWriter.write(self.field, 'index', self.indices, self.index_index)
-                self.index_index = 0
+                chars[char_index] = v
+                char_index += 1
+                if char_index == len(chars):
+                    DataWriter.write(self.field, 'values', chars, char_index)
+                    char_index = 0
+                accumulated += 1
+            indices[index_index] = accumulated
+            index_index += 1
+
+        if index_index > 0:
+            DataWriter.write(self.field, 'index', indices, index_index)
+        if char_index > 0:
+            DataWriter.write(self.field, 'values', chars, char_index)
+
 
     def flush(self):
-        if self.value_index != 0 or 'values' not in self.field:
-            DataWriter.write(self.field, 'values', self.values, self.value_index)
-            self.value_index = 0
-        if self.index_index != 0:
-            DataWriter.write(self.field, 'index', self.indices, self.index_index)
-            self.index_index = 0
+        # if self.value_index != 0 or 'values' not in self.field:
+        #     DataWriter.write(self.field, 'values', self.values, self.value_index)
+        #     self.value_index = 0
+        # if self.index_index != 0:
+        #     DataWriter.write(self.field, 'index', self.indices, self.index_index)
+        #     self.index_index = 0
         # self.field.attrs['fieldtype'] = self.fieldtype
         # self.field.attrs['timestamp'] = self.timestamp
         # self.field.attrs['chunksize'] = self.chunksize
@@ -285,13 +323,22 @@ class IndexedStringWriter(Writer):
             raise ValueError(f"'index' must be an ndarray of '{np.int64}'")
         if values.dtype not in (np.uint8, 'S1'):
             raise ValueError(f"'values' must be an ndarray of '{np.uint8}' or 'S1'")
-        DataWriter.write(self.field, 'index', index, len(index))
+        DataWriter.write(self.field, 'index', index[1:], len(index)-1)
         DataWriter.write(self.field, 'values', values, len(values))
 
     def write_raw(self, index, values):
         self.write_part_raw(index, values)
         self.flush()
 
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets, col_idx, written_row_count):
+        # broadcast accumulated size to current index array
+        index = column_inds[col_idx, :written_row_count + 1] + self.chunk_accumulated
+        self.chunk_accumulated += column_inds[col_idx, written_row_count]
+
+        col_offset = column_offsets[col_idx]
+        values = column_vals[col_offset: col_offset + column_inds[col_idx, written_row_count]]
+        self.write_part_raw(index, values)
+
 
 # TODO: should produce a warning for unmappable strings and a corresponding filter, rather
 # than raising an exception; or at least have a mode where this is possible
@@ -305,6 +352,8 @@ class LeakyCategoricalImporter:
         self.other_values = IndexedStringWriter(datastore, group, f"{name}_{out_of_range}",
                                                 timestamp, write_mode)
         self.field_size = max([len(k) for k in categories.keys()])
+        self.byte_map = ops.get_byte_map(categories)
+        self.freetext_index_accumulated = 0
 
     def chunk_factory(self, length):
         return np.zeros(length, dtype=f'U{self.field_size}')
@@ -335,6 +384,22 @@ class LeakyCategoricalImporter:
         self.write_part(values)
         self.flush()
 
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets, col_idx, written_row_count):
+        cat_keys, cat_index, cat_values = self.byte_map
+        chunk = np.zeros(written_row_count, dtype=np.int8) # use np.int8 instead of np.uint8, as we set -1 for leaky key
+        freetext_indices_chunk = np.zeros(written_row_count + 1, dtype = np.int64)
+
+        col_count = column_offsets[col_idx + 1] - column_offsets[col_idx]
+        freetext_values_chunk = np.zeros(np.int64(col_count), dtype = np.uint8)
+
+        ops.leaky_categorical_transform(chunk, freetext_indices_chunk, freetext_values_chunk, col_idx, column_inds, column_vals, column_offsets, cat_keys, cat_index, cat_values)
+
+        freetext_indices = freetext_indices_chunk + self.freetext_index_accumulated # broadcast
+        self.freetext_index_accumulated += freetext_indices_chunk[written_row_count]
+        freetext_values = freetext_values_chunk[:freetext_indices_chunk[written_row_count]]
+        self.writer.write_part(chunk)
+        self.other_values.write_part_raw(freetext_indices, freetext_values)
+
 
 # TODO: should produce a warning for unmappable strings and a corresponding filter, rather
 # than raising an exception; or at least have a mode where this is possible
@@ -347,15 +412,13 @@ class CategoricalImporter:
                                         categories, timestamp, write_mode)
         self.field_size = max([len(k) for k in categories.keys()])
 
+        self.byte_map = ops.get_byte_map(categories)
+
     def chunk_factory(self, length):
-        return np.zeros(length, dtype=f'U{self.field_size}')
+        return np.zeros(length, dtype='int8')
 
     def write_part(self, values):
-        results = np.zeros(len(values), dtype='int8')
-        keys = self.writer.keys
-        for i in range(len(values)):
-            results[i] = keys[values[i]]
-        self.writer.write_part(results)
+        self.writer.write_part(values)
 
     def flush(self):
         self.writer.flush()
@@ -364,6 +427,21 @@ class CategoricalImporter:
         self.write_part(values)
         self.flush()
 
+    def write_strings(self, values):
+        results = np.zeros(len(values), dtype='int8')
+        keys = self.writer.keys
+        for i in range(len(values)):
+            results[i] = keys[values[i]]
+        self.writer.write_part(results)
+        self.flush()
+
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets, col_idx, written_row_count):
+        chunk = np.zeros(written_row_count, dtype=np.uint8)
+        cat_keys, cat_index, cat_values = self.byte_map
+                
+        ops.categorical_transform(chunk, col_idx, column_inds, column_vals, column_offsets, cat_keys, cat_index, cat_values)
+        self.writer.write_part(chunk)
+
 
 class CategoricalWriter(Writer):
     def __init__(self, datastore, group, name, categories,
@@ -440,25 +518,62 @@ class NumericImporter:
         validity = np.zeros(len(values), dtype='bool')
         for i in range(len(values)):
             valid, value = self.parser(values[i], self.invalid_value)
-
             elements[i] = value
             validity[i] = valid
             
             if self.validation_mode == 'strict' and not valid: 
-                if self._is_blank_str(values[i]):
+                if self._is_blank(values[i]):
                     raise ValueError(f"Numeric value in the field '{self.field_name}' can not be empty in strict mode")  
                 else:        
                     raise ValueError(f"The following numeric value in the field '{self.field_name}' can not be parsed:{values[i].strip()}")
 
-            if self.validation_mode == 'allow_empty' and not self._is_blank_str(values[i]) and not valid:
-                raise ValueError(f"The following numeric value in the field '{self.field_name}' can not be parsed:{values[i].strip()}")
+            if self.validation_mode == 'allow_empty' and not self._is_blank(values[i]) and not valid:
+                raise ValueError(f"The following numeric value in the field '{self.field_name}' can not be parsed:{values[i]}")
 
         self.data_writer.write_part(elements)
         if self.flag_writer is not None:
             self.flag_writer.write_part(validity)
 
-    def _is_blank_str(self, value):
-        return type(value) == str and value.strip() == ''
+
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets, col_idx, written_row_count):
+        # elements = np.zeros(written_row_count, dtype=self.data_writer.nformat)
+        # validity = np.ones(written_row_count, dtype=bool)
+
+        value_dtype = ops.str_to_dtype(self.data_writer.nformat)
+
+        if self.data_writer.nformat == 'bool':
+            # TODO: replace with fast reader based on categorical string parsing
+            elements = np.zeros(written_row_count, dtype=self.data_writer.nformat)
+            validity = np.ones(written_row_count, dtype=bool)
+            exception_message, exception_args = ops.numeric_bool_transform(
+                elements, validity, column_inds, column_vals, column_offsets, col_idx,
+                written_row_count, self.invalid_value,
+                self.validation_mode, np.frombuffer(bytes(self.field_name, "utf-8"), dtype=np.uint8)
+            )
+        elif self.data_writer.nformat in ('int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'int64') :
+
+            exception_message, exception_args = 0, []
+            elements, validity = ops.transform_int_2(
+                column_inds, column_vals, column_offsets, col_idx,
+                written_row_count, self.invalid_value, self.validation_mode,
+                value_dtype, self.field_name)
+        else:
+            exception_message, exception_args = 0, []
+            elements, validity = ops.transform_float_2(
+                column_inds, column_vals, column_offsets, col_idx,
+                written_row_count, self.invalid_value, self.validation_mode,
+                value_dtype, self.field_name)
+
+        if exception_message != 0:
+            ops.raiseNumericException(exception_message, exception_args)
+
+        self.data_writer.write_part(elements)
+        if self.flag_writer is not None:
+            self.flag_writer.write_part(validity)
+
+
+    def _is_blank(self, value):
+        return (isinstance(value, str) and value.strip() == '') or value == b''
 
     def flush(self):
         self.data_writer.flush()
@@ -521,11 +636,17 @@ class FixedStringWriter(Writer):
         self.strlen = strlen
 
     def chunk_factory(self, length):
-        return np.zeros(length, dtype=f'S{self.strlen}')
+        return np.zeros(length, dtype='S{}'.format(self.strlen))
 
     def write_part(self, values):
         DataWriter.write(self.field, 'values', values, len(values))
 
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets,  col_idx, written_row_count):
+        values = np.zeros(written_row_count, dtype='S{}'.format(self.strlen))
+        ops.fixed_string_transform(column_inds, column_vals, column_offsets, col_idx,
+                                     written_row_count, self.strlen, values.data.cast('b'))
+        self.write_part(values)
+
     def flush(self):
         # self.field.attrs['fieldtype'] = self.fieldtype
         # self.field.attrs['timestamp'] = self.timestamp
@@ -594,6 +715,11 @@ class DateTimeImporter:
         self.write_part(values)
         self.flush()
 
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets, col_idx, written_row_count):
+        data = ops.transform_to_values(column_inds, column_vals, column_offsets, col_idx, written_row_count)
+        data = [x.tobytes().strip() for x in data]
+        self.write_part(data)
+
 
 # TODO writers can write out more than one field; offset could be done this way
 class DateTimeWriter(Writer):
@@ -760,6 +886,11 @@ class OptionalDateImporter:
             flags[i] = values[i] != b''
         return flags
 
+    def transform_and_write_part(self, column_inds, column_vals, column_offsets, col_idx, written_row_count):
+        data = ops.transform_to_values(column_inds, column_vals, column_offsets, col_idx, written_row_count)
+        data = [x.tobytes().strip() for x in data]
+        self.write_part(data)
+
     def flush(self):
         self.date.flush()
         if self.create_day_field:
@@ -770,3 +901,7 @@ class OptionalDateImporter:
     def write(self, values):
         self.write_part(values)
         self.flush()
+
+
+
+    
\ No newline at end of file
